#!/usr/bin/python
# Copyright 2016 CERN and GSI
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Description:
# This module implement the methods to build each components of the sources.
# A component source consists of one 'parameterizable' text that can
# be specialized on call to append the string to be stored in the file.
#

#=========================================================================
# Global definition
#=========================================================================

#-------------------------------------------------------------------------
# Hash-Table

import string

whichSimaticFormat = {
    'bool'    : 'BOOL',
	'uint8'   : 'BYTE',
	'int8'    : 'CHAR',
	'uint16'  : 'WORD',
	'int16'   : 'INT',
	'uint32'  : 'DWORD',
	'int32'   : 'DINT',
	'float32' : 'REAL',
    'string'   : 'STRING',
#	'uint64'  : '8',	not supported with PLC
#	'int64'   : '8',	not supported with PLC
#	'float64' : '8',    not supported with PLC
	'date'    : 'DT',
	'char'	:'CHAR',
    'byte'	:'BYTE',
    'word'	:'WORD',
    'dword'	:'DWORD',
    'int'	:'INT',
    'dint'	:'DINT',
    'real'	:'REAL',
    'dt'	:'DT',
    'dtl'	:'DTL'
}


head = string.Template("""\r
(* ---------------------------------------------------------------------\r
 * This file is auto generated by the SILECS framework tools. Code regeneration will overwrite it.\r
 * ---------------------------------------------------------------------\r
 *)\r
""")

#=========================================================================
# STL Source template (.scl file)
#=========================================================================
firstBlockUDT = string.Template("""\r
(* ---------------------------------------------------------------------\r
 * ${className}/ v${classVersion}\r
 * BLOCK Type definition\r
 * ---------------------------------------------------------------------\r
 *)\r
""")
                                
blockUDT = string.Template("""TYPE _${className}_${blockName}\r
AUTHOR:    ${owner}\r
FAMILY:    SILECS\r
NAME:      UDTB\r
    STRUCT\r
${regList}\r
    END_STRUCT;\r
END_TYPE\r
\r
""")

regScalar = string.Template("""        ${regName}: ${len}${value};\r
""")

regArray  = string.Template("""        ${regName}: ARRAY[0..${dim}] OF ${len}${value};\r
""")

regArray2d = string.Template("""        ${regName}: ARRAY[0..${dim1}, 0..${dim2}] OF ${len}${value};\r
""")

firstBlock = string.Template("""\r
(* ---------------------------------------------------------------------\r
 * ${className}/ v${classVersion}\r
 * BLOCK instance definition\r
 * ---------------------------------------------------------------------\r
 *)\r
""")

DB_TIAP = string.Template("""DATA_BLOCK ${className}_${blockName}${comment}\r
{ S7_Optimized_Access := 'FALSE' }\r
AUTHOR:    ${owner}\r
FAMILY:    SILECS\r
NAME:      ${mode}\r
STRUCT\r
${blockList}\r
END_STRUCT;\r
BEGIN\r
END_DATA_BLOCK\r
\r
""")

DB_STEP7 = string.Template("""(* ${className}_${blockName} ...........................................*)\r
DATA_BLOCK DB${DBnumber}${comment}\r
{ S7_Optimized_Access := 'FALSE' }\r
AUTHOR:    ${owner}\r
FAMILY:    SILECS\r
NAME:      ${mode}\r
STRUCT\r
${blockList}\r
END_STRUCT;\r
BEGIN\r
END_DATA_BLOCK\r
\r
""")

device = string.Template("""    ${label}: _${className}_${blockName};\r
""")

device_with_comment = string.Template("""    ${label}: _${className}_${blockName};  //${comment}\r
""")

#=========================================================================
# SYMBOLS Source template (.sdf file)
#=========================================================================

UDTSymbol  = string.Template(""""_${className}_${blockName}","UDT ${number}","UDT ${number}","${comment}"
""")

DBSymbol   = string.Template(""""${className}_${label}","DB ${number}","DB ${number}","${comment}"
""")

#=========================================================================
# Sub-function
#=========================================================================

# ------------------------------------------------------------------------
# Utils 
# ------------------------------------------------------------------------
def toCharArray(strg):
    charArray = ''
    for c in strg: charArray += "'%c'," %c
    return charArray

def toByteArray(strg):
    byteArray = ''
    for c in strg: byteArray += "16#%02x," %(ord(c))
    return byteArray

def toWordArray(strg):
    wordArray = ''
    for i in range(0, len(strg), 2):
        lsb = ord(strg[i])
        try: hsb = ord(strg[i+1]) 
        except: hsb = 0
        wordArray += "16#%02x%02x," %(hsb,lsb)
    return wordArray

# ------------------------------------------------------------------------
# STL code generation
# ------------------------------------------------------------------------

# REGISTER definition ----------------------------------------------------
def stlRegister(regName, regFormat, regDim, regDim2, regVal, regLen=1):
    if regLen > 1:    # register is a string or an array of strings
        strLen = whichSimaticFormat[regFormat]+'['+str(regLen)+']'
        if regDim == 1 and regDim2 == 1:    # scalar register
            return regScalar.substitute(regName=regName, len=strLen, value=regVal)
        elif regDim > 1 and regDim2 == 1:    # array register
            return regArray.substitute(regName=regName, dim=regDim-1, len=strLen, value=regVal)
        else:    # dim1>=1 and for whatever dim2, use double array syntax
            return regArray2d.substitute(regName=regName, dim1=regDim-1, dim2=regDim2-1, len=strLen, value=regVal)
    else:
        if regDim == 1 and regDim2 == 1:
            return regScalar.substitute(regName=regName, len=whichSimaticFormat[regFormat], value=regVal)
        elif regDim > 1 and regDim2 == 1:
            return regArray.substitute(regName=regName, dim=regDim-1, len=whichSimaticFormat[regFormat], value=regVal)
        else:    # dim1>=1 and for whatever dim2, use double array syntax
            return regArray2d.substitute(regName=regName, dim1=regDim-1, dim2=regDim2-1, len=whichSimaticFormat[regFormat], value=regVal)

#This method is used to add default value to the diagnostic registers
def stlRegisterValue(regName, owner, date, checksum, silecsversion):
	if regName == '_version': return " := 'SILECS_%s'" %silecsversion
	if regName == '_checksum' : return " := DW#16#%x" %checksum
	if regName == '_user'   : return " := '%s'" %owner
	if regName == '_date'   : return " := %s" %date
	return ''    #all other registers cannot have default assignment for the time being

# BLOCK type definition (UDT) --------------------------------------------
def stlBlockUDT(owner, className, classVersion, blockName, regList, isFirst):
    if isFirst:
        return firstBlockUDT.substitute(className=className, classVersion=classVersion) + blockUDT.substitute(className=className, blockName=blockName, owner=owner, regList=regList)
    else:
        return blockUDT.substitute(className=className, blockName=blockName, owner=owner, regList=regList)
    
# BLOCK_MODE: data-block instance ----------------------------------------
def stlDevice(devLabel, className, blockName, fesaName):
    if fesaName == "":
        return device.substitute(label=devLabel, className=className, blockName=blockName)
    return device_with_comment.substitute(label=devLabel, className=className, blockName=blockName, comment=fesaName)
	    
# DEVICE_MODE: data-block instance ---------------------------------------
def stlBlock(className, blockName):
    return device.substitute(label=blockName, className=className, blockName=blockName)     
        
def generateBlock(owner, className, classVersion, system, DBnumber, blockName, blockList, isFirst, protocol, fesaName):
    srcCore = ""
    mode = ''
    comment = ''
    if protocol == 'BLOCK_MODE':
        mode = 'BLK_MODE'
    elif protocol == 'DEVICE_MODE':
        mode = 'DEV_MODE'
        if fesaName != "":
            comment = "  //" + fesaName
    else:
        raise Exception( "Unknown PLC-protocol defined: " + protocol )

    if (system == 'TIA-PORTAL'):
        srcCore = DB_TIAP.substitute(className=className, blockName=blockName, owner=owner, mode=mode, blockList=blockList, comment=comment)
    elif (system == 'STEP-7'):
        srcCore = DB_STEP7.substitute(className=className, blockName=blockName, DBnumber=DBnumber, owner=owner, mode=mode, blockList=blockList, comment=comment)
    else:
        raise Exception( "Unknown PLC-system defined: " + system )

    if isFirst:
        return firstBlock.substitute(className=className, classVersion=classVersion) + srcCore
    else:
        return srcCore
	
# ------------------------------------------------------------------------
# SYMBOLS code generation
# ------------------------------------------------------------------------

# HEADER file ------------------------------------------------------------
def symHeader(baseAddress):
    return DBSymbol %('SILECS', 'HEADER', baseAddress, baseAddress, "SILECS Diagnostic data-block")

# UDT Type definition ----------------------------------------------------
def symBlockUDT(className, classVersion, blockName, number, isFirst):
    comment = ''
    if isFirst: comment = "[%s/%s] UDT symbol: _<class-name>_<block-name>" %(className, classVersion)
    return UDTSymbol.substitute(className=className, blockName=blockName, number=number, comment=comment)

# BLOCK_MODE: data-block symbol ----------------------------
def symBlockDB(className, classVersion, blockName, number, isFirst):
    comment = ''
    if isFirst: comment = "[%s/%s] DB symbol: <class-name>_<block-name>" %(className, classVersion)
    return DBSymbol.substitute(className=className, label=blockName, number=number, comment=comment)

# DEVICE_MODE : data-block symbol ----------------------------
def symDeviceDB(className, classVersion, deviceLabel, number, isFirst):
    comment = ''
    if isFirst: comment = "[%s/%s] DB symbol: <class-name>_<device-label | device-id>" %(className, classVersion)
    return DBSymbol.substitute(className=className, label=deviceLabel, number=number, comment=comment)
